"""Any HDRezka page"""
__all__ = ('Page',)

from typing import Iterable

from bs4 import BeautifulSoup

from .inline import *
from .._bs4 import BUILDER
from ..api.http import get_response
from ..errors import EmptyPage
from ..url import Request


class Page:
    """AJAX class for HDRezka search"""
    __slots__ = ('_page', '_page_format', '__yields', '__yields_page')

    def __init__(self, url: str = Request.HOST):
        self.__yields: list[InlineItem] = []
        self.__yields_page = 0
        self.page = url

    @property
    def page(self) -> str:
        """Current page"""
        return self._page

    @page.setter
    def page(self, value):
        """Cast value to str and sets"""
        if not isinstance(value, str):
            value = str(value)
        # noinspection HttpUrlsUsage
        if not (value.startswith('https://') or value.startswith('http://')):
            value = f'https://{value}'
        self._page = value
        self._page_format = self._concat_paginator(value.removesuffix('/')).format

    @staticmethod
    def _concat_paginator(url: str) -> str:
        return f'{url}/page/{{0}}'

    @staticmethod
    def _inline_info(years: str, country: str, genre: str):
        year, *finals = years.split('-')
        if finals:
            final, = finals
            year_final = ... if final.strip() == '...' else int(final)
        else:
            year_final = None
        return InlineInfo(int(year), year_final, country.strip(), genre.strip())

    async def get_page(self, page: int | slice | Iterable[int]) -> list[InlineItem]:
        """
        Get items from pages range.
        None for all elements of all pages
        """
        content = (await get_response('GET', self._page_format(page))).text
        items = BeautifulSoup(content, builder=BUILDER
                              ).find_all(class_='b-content__inline_item')
        if not items:
            raise EmptyPage(page)
        return [InlineItem(
            (a := (link := i.find(class_='b-content__inline_item-link')).find('a'))['href'],
            a.text,
            self._inline_info(*link.find('div').text.split(',', 3)),
            i.find(class_='b-content__inline_item-cover').find('img').get('src', ''))
            for i in items]

    def __aiter__(self):
        """Async iterator by pages"""
        self.__yields_page = 0
        self.__yields.clear()
        return self

    async def __anext__(self) -> InlineItem:
        """Returns `InlineItem` object for every title in page"""
        if not self.__yields:
            try:
                self.__yields_page += 1
                self.__yields += await self.get_page(self.__yields_page)
            except EmptyPage:
                raise StopAsyncIteration
        return self.__yields.pop(0)

    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.page!r})'



"""Subtitles module"""
__all__ = ('SubtitleURL', 'SubtitleURLs')

from typing import NamedTuple


class SubtitleURL(NamedTuple):
    """
    url: str
        .vtt file url

    Language attributes:
    name: str
    code: str
    """
    url: str
    name: str
    code: str


class SubtitleURLs:
    """Class representing subtitle urls"""
    __slots__ = ('subtitles', 'has_subtitles', 'subtitle_names', 'subtitle_codes', 'default')

    def __init__(self, subtitle: str, subtitle_lns: dict[str, str], subtitle_def: str):
        """
        :param subtitle: is subtitles exists
        :param subtitle_lns: languages {code: name, ...}
        :param subtitle_def: default subtitle code
        """
        self.has_subtitles = not not subtitle
        off = SubtitleURL('', '', 'off'),
        if self.has_subtitles:
            self.subtitles: tuple[SubtitleURL, ...] = *(
                SubtitleURL(url, name, subtitle_lns[name]) for name, url in
                (v.removeprefix('[').split(']', 1) for v in subtitle.split(','))),
            self.subtitles += off
        else:
            self.subtitles = off
        self.subtitle_names = {}
        self.subtitle_codes = {}
        for subtitle_item in self.subtitles:
            self.subtitle_names[subtitle_item.name] = self.subtitle_codes[subtitle_item.code] = subtitle_item
        self.default: SubtitleURL | None = self.subtitle_codes.get(subtitle_def)

    def __getitem__(self, item: str) -> SubtitleURL:
        """Returns subtitle by name or code"""
        if item in self.subtitle_names:
            return self.subtitle_names[item]
        return self.subtitle_codes[item]

    def get(self, item: str) -> SubtitleURL | None:
        """Returns subtitle by name or code, if not found returns None"""
        if not self.has_subtitles:
            return None
        return self.subtitle_names.get(item) or self.subtitle_codes.get(item)

    def __getattr__(self, item: str) -> SubtitleURL:
        """Returns subtitle by code"""
        return self.subtitle_codes[item]

    def __bool__(self):
        """Is subtitles exists"""
        return self.has_subtitles

    def __repr__(self):
        return f'{self.__class__.__qualname__}<{self.subtitles!r}>'



"""High-level structures for representing title data"""
__all__ = ('Hyperlink', 'Rating', 'Rank', 'Release', 'AgeRating', 'Poster', 'Birthplace')

from typing import NamedTuple


class Hyperlink(NamedTuple):
    """Hyperlink representation"""
    name: str = ''
    url: str = ''

    def __str__(self):
        return self.name


class Rating(NamedTuple):
    """Rating for service"""
    service: Hyperlink
    rating: int | float
    votes: int


class Rank(NamedTuple):
    """Rank of title"""
    name: Hyperlink
    rank: int


class Release(NamedTuple):
    """Release date"""
    day: str
    year: int


class AgeRating(NamedTuple):
    """Age restricts"""
    age: int
    description: str = ''


class Poster(NamedTuple):
    """Poster of title"""
    full: str = ''
    preview: str = ''

    def __bool__(self):
        """Is any poster"""
        return not not (self.full or self.preview)


class Birthplace(NamedTuple):
    """Birthplace information"""
    country: str
    city: str | None = None
    subcountry: str | None = None
    state: str | None = None



"""HDRezka deobfuscation process"""
__all__ = ('clear_trash',)

import re
from binascii import a2b_base64

_sub_trash = re.compile(
    '^#h|//_//|'
    '(?:I[01UV]|[JQ][EF]|X[kl])(?:[A4][=hjk]|[B5][Ae])|'
    '(?:I[Sy]|[JQ]C|Xi)(?:[EMQ][=hjk]|[FNR][Ae])'
).sub


def clear_trash(trash_string: str) -> str:
    """removes base64 trash from string"""
    return a2b_base64(_sub_trash('', trash_string).encode('ASCII') + b'==').decode('ASCII')



"""Presentation of title information"""
__all__ = ('PostInfo',)

from typing import NamedTuple

from .fields import *
from .person import Person
from ...post.urls.kind.quality import Quality


class PostInfo(NamedTuple):
    """General post information"""
    title: str
    orig_title: str
    poster: Poster
    duration: int
    updated_time: int
    ratings: dict[str, Rating]
    rankings: tuple[Rank, ...]
    slogan: str
    release: Release
    country: tuple[Hyperlink, ...]
    director: str
    genre: tuple[Hyperlink, ...]
    quality: Quality | None
    translators: tuple[str, ...]
    age_rating: AgeRating
    duration_m: int
    collections: tuple[Hyperlink, ...]
    persons: tuple[Person, ...]
    description: str



"""All urls for one video"""
__all__ = ('URLs', 'urls_from_ajax_response')

from typing import NamedTuple

from .kind import SubtitleURLs, VideoURLs
from ...api.types import APIResponse


class URLs(NamedTuple):
    """All urls for one video class"""
    video: VideoURLs
    subtitles: SubtitleURLs


def urls_from_ajax_response(response: APIResponse) -> URLs:
    """Parse response and return urls"""
    return URLs(VideoURLs(response.get('url', '')),
                SubtitleURLs(response.get('subtitle', ''),
                             response.get('subtitle_lns', {'off': ''}),
                             response.get('subtitle_def', '')))



"""Global request settings"""
__all__ = ('Request',)

from urllib.parse import urljoin


class Request:
    """
    Contains global request parameters for HDRezka.
    Change `HOST` here in case the existing host is unavailable.

    See also [hdrezka.api.http.DEFAULT_CLIENT]
    """
    HOST: str = 'https://hdrezka.me/'
    REDIRECT_URL: str = 'https://rhs.to/'
    'Used in `hdrezka.api.http.login_global`'

    @classmethod
    def host_join(cls, url: str | None, allow_fragments: bool = True) -> str:
        """urljoin with `cls.HOST`"""
        return urljoin(cls.HOST, url, allow_fragments=allow_fragments)
